import re
from logging import LogRecord
from typing import Optional

from PySide6.QtCore import Slot
from PySide6.QtWidgets import QGroupBox, QVBoxLayout, QWidget, QPlainTextEdit, QLabel, QLineEdit, QSizePolicy
from tinyioc import inject

from ..palette.base import BaseColorPalette

from .. import constants


class TaskPanel(QGroupBox):
    _layout: QVBoxLayout
    _actions_group_widget: QWidget
    _actions_group_layout: QVBoxLayout
    _output_stream_edit_text: QPlainTextEdit

    _output_stream: str

    _task_name: str
    _task_label: QLabel

    _filter_edit_text: QLineEdit

    @inject()
    def __init__(self, parent: Optional[QWidget], task_name: str, palette: BaseColorPalette):
        super(TaskPanel, self).__init__(parent)

        self._task_name = task_name
        self._output_stream = ""

        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)

        self.setStyleSheet(f"""
            background-color: {palette.current_line};
            border: 0;
        """)

        self._actions_group_widget = QWidget(self)
        self._actions_group_layout = QVBoxLayout(self)
        self._actions_group_layout.setContentsMargins(0, 0, 0, 0)
        self._actions_group_widget.setLayout(self._actions_group_layout)

        self._layout.addWidget(self._actions_group_widget)

        self._task_label = QLabel(self)
        self._task_label.setText(self._task_name)
        self._task_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self._task_label.setStyleSheet(f"""
            color: {palette.foreground};
            font-weight: bold;
        """)
        self._actions_group_layout.addWidget(self._task_label)

        self._filter_edit_text = QLineEdit(self)
        self._filter_edit_text.setPlaceholderText("Filter")
        self._filter_edit_text.setStyleSheet(f"""
            background-color: {palette.background};
            color: {palette.foreground};
        """)
        self._filter_edit_text.textChanged.connect(self._filter_changed)
        self._actions_group_layout.addWidget(self._filter_edit_text)

        self._output_stream_edit_text = QPlainTextEdit(self)
        self._output_stream_edit_text.setReadOnly(True)
        self._output_stream_edit_text.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self._output_stream_edit_text.setStyleSheet(f"""
            background-color: {palette.background}; 
            color: {palette.foreground}; 
            font-family: monospace;
        """)
        self._layout.addWidget(self._output_stream_edit_text, 1)

    # noinspection PyUnresolvedReferences
    def append_text(self, record: LogRecord):
        scroll_bottom = False
        previous_scrollbar_pos = self._output_stream_edit_text.verticalScrollBar().value()

        if self._output_stream_edit_text.verticalScrollBar().value() > \
                self._output_stream_edit_text.verticalScrollBar().maximum() - constants.SCROLL_TOLERANCE:
            scroll_bottom = True

        processed_message = re.sub(r'\x1b\[([0-9,A-Z]{1,2}(;[0-9]{1,2})?(;[0-9]{3})?)?[m|K]?', '', record.message)
        self._output_stream += processed_message
        self._update_output_edit_text()

        if scroll_bottom:
            self._output_stream_edit_text.verticalScrollBar().setValue(
                self._output_stream_edit_text.verticalScrollBar().maximum())
        else:
            self._output_stream_edit_text.verticalScrollBar().setValue(
                min(self._output_stream_edit_text.verticalScrollBar().maximum(),
                    max(self._output_stream_edit_text.verticalScrollBar().minimum(), previous_scrollbar_pos)))

    def _update_output_edit_text(self):
        filter_input = self._filter_edit_text.text()

        if filter_input:
            filtered_text = ""
            for line in self._output_stream.splitlines(keepends=True):
                if filter_input in line:
                    filtered_text += line
        else:
            filtered_text = self._output_stream

        self._output_stream_edit_text.setPlainText(filtered_text)

    @Slot()
    def _filter_changed(self):
        self._update_output_edit_text()



#!/usr/bin/env python
import argparse
import sys
import traceback
from logging.handlers import QueueHandler
from multiprocessing import Queue
from tinyioc import register_instance

from jorun.runner_process import RunnerProcess
from .palette.base import BaseColorPalette
from .palette.darcula import DarculaColorPalette
from .ui.application import UiApplication
from .handler.group import GroupTaskHandler
from .handler.docker import DockerTaskHandler
from .handler.shell import ShellTaskHandler

from .configuration import load_config, AppConfiguration
from .types.task import TasksConfiguration
from .logger import logger, NewlineStreamHandler

parser = argparse.ArgumentParser(prog="jorun", description="A smart task runner", add_help=True)

parser.add_argument("configuration_file", help="The yml configuration file to run")
parser.add_argument("--level", help="The log level (DEBUG, INFO, ...)", default="INFO", type=str)
parser.add_argument("--file-output", help="Log tasks output to files, one per task. "
                                          "This option lets you specify the directory of the log files", type=str)
parser.add_argument("--gui", help="Force running with the graphical interface", action='store_true')
parser.add_argument("--no-gui", help="Force running without the graphical interface", action='store_true')

program_arguments: argparse.Namespace

task_streams_queue = Queue()

ui_application: UiApplication


def main():
    global program_arguments, ui_application

    register_instance(DarculaColorPalette(), register_for=BaseColorPalette)

    program_arguments = parser.parse_args()
    logger.setLevel(program_arguments.level)

    logger.debug("Loading configuration file")
    config: TasksConfiguration = load_config(program_arguments.configuration_file)

    missing_tasks = config["tasks"].copy()

    tasks_config = config.get("tasks")

    if not tasks_config:
        raise RuntimeError("No tasks to run found")

    gui_config = config.get("gui")
    show_gui = not program_arguments.no_gui and (program_arguments.gui or gui_config)

    if show_gui:
        logger.debug("Using graphical interface")
        log_handler = QueueHandler(task_streams_queue)
    else:
        logger.debug("Using console output")
        log_handler = NewlineStreamHandler(sys.stdout)

    tasks_thread = RunnerProcess(tasks_config, program_arguments, show_gui, task_streams_queue)
    tasks_thread.start()

    try:
        if show_gui:
            ui_tasks = [t_name for t_name, t_val in missing_tasks.items() if t_val["type"] != "group"]

            ui_application = UiApplication(ui_tasks, task_streams_queue, gui_config)
            ui_application.start_ui()
        else:
            tasks_thread.join()
    except KeyboardInterrupt:
        logger.debug("Requested termination")
    except Exception as e:
        logger.error("An error occurred")
        traceback.print_exception(e)
    finally:
        if show_gui:
            logger.debug("Quitting the UI")
            ui_application.stop_ui()
        logger.debug("Quitting the tasks")
        tasks_thread.stop(10)

    logger.debug("Terminated")


if __name__ == "__main__":
    main()



from PySide6.QtCore import QObject, Signal


class MainWindowSignals(QObject):
    data_received = Signal(object)
    app_terminated = Signal()


class DataUpdateSignalEmitter:
    signals: MainWindowSignals



import multiprocessing
import queue
import sys
from logging.handlers import QueueHandler
from multiprocessing.connection import Connection
from threading import Thread
from typing import List, Set, Dict, Optional
import asyncio
import logging
import traceback

from tinyioc import module, IocModule, register_instance, unregister_service

from .configuration import AppConfiguration
from .handler.docker import DockerTaskHandler
from .handler.group import GroupTaskHandler
from .handler.shell import ShellTaskHandler
from .types.task import Task
from .runner import TaskRunner
from .logger import logger, NewlineStreamHandler


@module()
class RunnerThreadModule(IocModule):
    pass


class RunnerProcess(multiprocessing.Process):
    _running: bool
    _pipe_emit: Connection
    _pipe_recv: Connection

    _config: Dict[str, Task]
    _arguments: any
    _running_tasks: List[TaskRunner]
    _async_tasks: Set[asyncio.Task]

    _missing_tasks: Dict[str, Task]
    _completed_tasks: Set[str]

    _queue: multiprocessing.Queue
    _log_handler: logging.Handler
    _show_gui: bool

    _loop: asyncio.AbstractEventLoop

    def __init__(self, configuration: Dict[str, Task], arguments: any, is_gui: bool,
                 output_queue: Optional[multiprocessing.Queue]):
        super(RunnerProcess, self).__init__()

        self._show_gui = is_gui
        self._config = configuration
        self._arguments = arguments
        self._queue = output_queue

        self._pipe_recv, self._pipe_emit = multiprocessing.Pipe()

    def _run_missing_tasks(self):
        tasks_to_run: List[Task] = [t for t in self._missing_tasks.values() if
                                    (not t.get("depends") or len(t["depends"]) == 0) or set(t["depends"]).issubset(
                                        self._completed_tasks)]

        for task in tasks_to_run:
            self._missing_tasks.pop(task["name"])

        for task in tasks_to_run:
            self._run_task(task)

    def _run_task(self, task: Task):
        t = TaskRunner(task, self._arguments.file_output, self._arguments.level,
                       self._log_handler)
        self._running_tasks.append(t)

        def cb():
            logger.debug(f"Task {task['name']} completed")
            self._completed_tasks.add(task["name"])

            logger.debug(f"Launching task {task['name']} dependencies")
            self._run_missing_tasks()

        logger.debug(f"Running task {task['name']}")
        async_t = self._loop.create_task(t.start(cb))
        self._async_tasks.add(async_t)
        async_t.add_done_callback(self._async_tasks.discard)

    def _cancel_tasks(self):
        logger.debug("Killing running tasks...")
        for i in reversed(range(len(self._running_tasks))):
            t = self._running_tasks[i]
            logger.debug(f"Killing task {t.name}")
            try:
                t.stop()
            except:
                pass
            finally:
                self._running_tasks.pop(i)

    def _cancel_async_tasks(self):
        logger.debug("Killing async tasks...")
        for t in self._async_tasks.copy():
            t.cancel()

    def run(self) -> None:
        register_instance(AppConfiguration([
            ShellTaskHandler(),
            DockerTaskHandler(),
            GroupTaskHandler()
        ]))

        self._log_handler = QueueHandler(self._queue) if self._show_gui else NewlineStreamHandler(sys.stdout)
        self._running_tasks = []
        self._async_tasks = set()
        self._missing_tasks = self._config.copy()
        self._completed_tasks = set()

        self._running = True

        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        register_instance(self._loop, module=RunnerThreadModule, register_for=asyncio.AbstractEventLoop)

        async def periodic_termination_checker():
            while self._running:
                if self._pipe_recv.poll():
                    self._loop.call_soon_threadsafe(self._loop.stop)
                    break

                await asyncio.sleep(0.3)

        try:
            self._run_missing_tasks()

            term_task = self._loop.create_task(periodic_termination_checker())
            self._async_tasks.add(term_task)
            term_task.add_done_callback(self._async_tasks.discard)

            self._loop.run_forever()
        except KeyboardInterrupt:
            logger.info("Requested termination")
        except Exception as e:
            logger.error("An error occurred")
            traceback.print_exception(e)
        finally:
            unregister_service(asyncio.AbstractEventLoop, module=RunnerThreadModule)

            self._cancel_async_tasks()
            self._cancel_tasks()

            if self._loop.is_running():
                logger.debug("Terminating the async loop...")
                self._loop.stop()

            logger.debug("Closing the async loop...")
            self._loop.close()

        self._running = False

    def stop(self, timeout: Optional[float] = None):
        self._pipe_emit.send(1)
        try:
            self.join(timeout)
        except:
            logger.debug("Task executor process terminated abruptly")
        logger.debug("Tasks process terminated")
